/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Files de priorité, réalisées par un arbre binaire complet dans un tableau

// La capacité est ici limitée, mais il serait facile de remplacer le
// tableau par un tableau redimensionnable (cf vector.h)

#include "prelude.h"

#include "pqueue.h"

struct Pqueue {
  int capacity;
  int size;     // 0 <= size <= capacity
  int *data;    // tableau de taille capacity
};

pqueue *pqueue_create(int capacity) {
  assert(capacity >= 0);
  pqueue *q = malloc(sizeof(struct Pqueue));
  q->capacity = capacity;
  q->data = calloc(capacity, sizeof(int));
  q->size = 0;
  return q;
}

int pqueue_size(pqueue *q) {
  return q->size;
}

bool pqueue_is_empty(pqueue *q) {
  return q->size == 0;
}

bool pqueue_is_full(pqueue *q) {
  return q->size == q->capacity;
}

void move_up(int a[], int i, int x) {
  while (i > 0) {
    int fi = (i - 1) / 2;
    int y = a[fi];
    if (y <= x) break;
    a[i] = y;
    i = fi;
  }
  a[i] = x;
}

void pqueue_add(pqueue *q, int x) {
  assert(!pqueue_is_full(q));
  move_up(q->data, q->size, x);
  q->size++;
}

int pqueue_min(pqueue *q) {
  assert(!pqueue_is_empty(q));
  return q->data[0];
}

void move_down(int a[], int i, int x, int n) {
  while (true) {
    int j = 2 * i + 1;
    if (j >= n) break;
    if (j + 1 < n && a[j] > a[j + 1]) { j++; }
    if (a[j] >= x) break;
    a[i] = a[j];
    i = j;
  }
  a[i] = x;
}

int pqueue_remove_min(pqueue *q) {
  assert(!pqueue_is_empty(q));
  int r = q->data[0];
  int n = --q->size;
  if (n > 0) move_down(q->data, 0, q->data[n], n);
  return r;
}

void pqueue_delete(pqueue *q) {
  free(q->data);
  free(q);
}

pqueue *pqueue_of_array(int a[], int n) {
  pqueue *q = pqueue_create(n);
  q->size = n;
  for (int i = n / 2; i < n; i++) {
    q->data[i] = a[i];
  }
  for (int i = n / 2 - 1; i >= 0; i--) {
    move_down(q->data, i, a[i], n);
  }
  return q;
}
